package org.qatools.validator;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import org.apache.log4j.Logger;


/**
 *
 */
public class StackUnwinder {
    private static Logger logger = Logger.getLogger(StackUnwinder.class);
    public static boolean logInfo = false;

    private static final String TEST_METHOD_NAME_PREFIX = "test";
    private static final String TEST_METHOD_NAME_PREFIX2 = "setUp";
    private static final String TEST_METHOD_NAME_PREFIX3 = "oneTimeSetup";
                            
    private static final String JUNIT_ANNOTATION_NAME = "org.junit.Test";
    private static final String JUNIT_ANNOTATION_NAME2 = "org.junit.Before";
    private static final String JUNIT_ANNOTATION_NAME3 = "org.junit.BeforeClass";
    private static final String JUNIT_ANNOTATION_NAME4 = "org.junit.After";
    private static final String JUNIT_ANNOTATION_NAME5 = "org.junit.AfterClass";
                            
    private static final String TESTNG_ANNOTATION_NAME = "org.testng.annotations.Test";
    private static final String TESTNG_ANNOTATION_NAME2 = "org.testng.annotations.BeforeClass";
    private static final String TESTNG_ANNOTATION_NAME3 = "org.testng.annotations.BeforeGroups";
    private static final String TESTNG_ANNOTATION_NAME4 = "org.testng.annotations.BeforeMethod";
    private static final String TESTNG_ANNOTATION_NAME5 = "org.testng.annotations.BeforeSuite";
    private static final String TESTNG_ANNOTATION_NAME6 = "org.testng.annotations.BeforeTest";
    private static final String TESTNG_ANNOTATION_NAME7 = "org.testng.annotations.AfterClass";
    private static final String TESTNG_ANNOTATION_NAME8 = "org.testng.annotations.AfterGroups";
    private static final String TESTNG_ANNOTATION_NAME9 = "org.testng.annotations.AfterMethod";
    private static final String TESTNG_ANNOTATION_NAME10 = "org.testng.annotations.AfterSuite";
    private static final String TESTNG_ANNOTATION_NAME11 = "org.testng.annotations.AfterTest";
                            
    //
    // for testcase stack
    private static final String DEFAULT_TC_PACKAGE = "default.testcase.package";
    private static final String DEFAULT_TC_SUITE = "defaultSuite";
    private static final String DEFAULT_TC_NAME = "defaultTestcaseName";
    private static final String DEFAULT_TC_LINE = "-1";
    
    //
    // for actual calling method stack
    private static final String DEFAULT_ACTUAL_PACKAGE = "default.actual.package";
    private static final String DEFAULT_ACTUAL_NAME = "defaultActualMethodName";
    private static final String DEFAULT_ACTUAL_LINE = "-1";

    
    private StackUnwinder() {
    }


    /**
     * Returns information about the current TestCase
     *
     *     	stackDetails[0] = DEFAULT_TC_PACKAGE;
     *    	stackDetails[1] = DEFAULT_TC_SUITE;
     *    	stackDetails[2] = DEFAULT_TC_NAME;
     *     	stackDetails[3] = DEFAULT_TC_LINE;
     *     
     */
    public static String[] getCallerTestMethodDetails() {
    	String[] stackDetails = new String[4];
    	stackDetails[0] = DEFAULT_TC_PACKAGE;
    	stackDetails[1] = DEFAULT_TC_SUITE;
    	stackDetails[2] = DEFAULT_TC_NAME;
    	stackDetails[3] = DEFAULT_TC_LINE;
    	
        StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
        for (StackTraceElement stackTraceElement : stackTrace) {
            if (isTestMethodByAnnotation(stackTraceElement) || isTestMethodByName(stackTraceElement)) {
                //log.debug("Detected method with @Test annotation");            	
    			String[] parts = stackTraceElement.getClassName().split("\\.");
    			String suite = parts[parts.length-1];
    			String pack = stackTraceElement.getClassName().replace("."+suite, "");
    			String testCaseName = stackTraceElement.getMethodName();
    			String linenum = "" + stackTraceElement.getLineNumber();

    	    	stackDetails[0] = pack;
    	    	stackDetails[1] = suite;
    	    	stackDetails[2] = testCaseName;
    	    	stackDetails[3] = linenum;
                return stackDetails;
            }
        }
        return stackDetails;
    }

    /**
     * Returns information about the actual calling method
     *
     *     	stackDetails[0] = DEFAULT_ACTUAL_PACKAGE;
     *    	stackDetails[1] = DEFAULT_ACTUAL_NAME;
     *    	stackDetails[2] = DEFAULT_TC_LINE;
     *     
     */
    public static String[] getActualCallerDetails() {
		String[] stackDetails = new String[3];
		stackDetails[0] = DEFAULT_ACTUAL_PACKAGE;
    	stackDetails[1] = DEFAULT_ACTUAL_NAME;
    	stackDetails[2] = DEFAULT_TC_LINE;

        StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
		String actualCallingClass = stackTrace[3].getClassName();
		String actualCallingMethod = stackTrace[3].getMethodName();
		String actualCallingLine = "" + stackTrace[3].getLineNumber();

		stackDetails[0] = actualCallingClass;
    	stackDetails[1] = actualCallingMethod;
    	stackDetails[2] = actualCallingLine;

        return stackDetails;

    }

    /**
     * Checks if stack trace element is test case using expected JUnit annotation @Test
     *
     * @param stackTraceElement stack trace element to analyze
     * @return true if stack trace element is test method
     */
    private static boolean isTestMethodByAnnotation(StackTraceElement stackTraceElement) {
        String className = stackTraceElement.getClassName();
        String methodName = stackTraceElement.getMethodName();
        try {
            // getting class
            Class<?> clazz = Class.forName(className);
            //getting methods
            Method[] methods = clazz.getMethods();
            for (Method method : methods) {
                method.setAccessible(true);
                // there can't be 2 methods with the same name in test class as test methods don't have parameters
                if (method.getName().equals(methodName)) {
                    // getting annotations
                    Annotation[] annotations = method.getAnnotations();
                    for (Annotation annotation : annotations) {
                        // checking annotation
                        String annotationName = annotation.annotationType().getName();
                        if (    JUNIT_ANNOTATION_NAME.equals(annotationName)        // Junit annotations
                                || JUNIT_ANNOTATION_NAME2.equals(annotationName)
                                || JUNIT_ANNOTATION_NAME3.equals(annotationName)
                                || JUNIT_ANNOTATION_NAME4.equals(annotationName)
                                || JUNIT_ANNOTATION_NAME5.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME.equals(annotationName)    //TestNG annotations
                                || TESTNG_ANNOTATION_NAME2.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME3.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME4.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME5.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME6.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME7.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME8.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME9.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME10.equals(annotationName)
                                || TESTNG_ANNOTATION_NAME11.equals(annotationName)
                                ) {
                            return true;
                        }
                    }
                }
            }
            return false;
        } catch (Exception e) {
            String msg = "Failed to detect method annotations. Error: " + e;
            //log.error(msg, e);
            //todo: maybe exception should be re-thrown here instead
            return false;
        }
    }

    /**
     * Checks if stack trace element is test case using expected name prefix
     *
     * @param stackTraceElement stack trace element to analyze
     * @return true if stack trace element is test method
     */
    private static boolean isTestMethodByName(StackTraceElement stackTraceElement) {
        String methodName = stackTraceElement.getMethodName();
        return  methodName.startsWith(TEST_METHOD_NAME_PREFIX) ||
                    methodName.startsWith(TEST_METHOD_NAME_PREFIX2) ||
                    methodName.startsWith(TEST_METHOD_NAME_PREFIX3);
    }

    /**
     * 
     * @param stackDetail
     */
    public static void log(String[] stackDetails) {
        if (!logInfo) {
            return;
        }
        
        if (stackDetails == null || stackDetails.length == 0) {
            return;
        }
        
        String s = "";
        for (int i=0; i<stackDetails.length; i++) {
            s += stackDetails[i] + ", ";
        }
        logger.info (s);
    }
    
}

